from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_search._private.PluginNames import searchTuplePrefix
from peek_core_search._private.storage.DeclarativeBase import DeclarativeBase


@addTupleType
class ExcludeSearchStringTable(DeclarativeBase, Tuple):
    __tupleType__ = searchTuplePrefix + "ExcludeSearchStringTable"
    __tablename__ = "ExcludeSearchString"

    id = Column(Integer, primary_key=True, autoincrement=True)
    term = Column(String, nullable=False)
    partial = Column(Boolean, nullable=True)
    full = Column(Boolean, nullable=True)
    comment = Column(String, nullable=True)

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_search._private.PluginNames import searchTuplePrefix
from peek_core_search._private.storage.DeclarativeBase import DeclarativeBase


@addTupleType
class SearchObjectTypeTuple(DeclarativeBase, Tuple):
    __tupleType__ = searchTuplePrefix + "SearchObjectTypeTuple"
    __tablename__ = "SearchObjectType"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    title = Column(String, nullable=False)
    order = Column(Integer, nullable=False, server_default="0")

    __table_args__ = (
        Index("idx_ObjType_name", name, unique=True),
        Index("idx_ObjType_title", title, unique=True),
    )

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



import logging
from pathlib import Path
from typing import List, Any

from peek_abstract_chunked_index.private.client.controller.ACICacheControllerABC import (
    ACICacheControllerABC,
)
from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.server.client_handlers.ClientChunkLoadRpc import (
    ClientChunkLoadRpc,
)
from peek_core_search._private.storage.EncodedSearchIndexChunk import (
    EncodedSearchIndexChunk,
)
from peek_core_search._private.tuples.search_index.SearchIndexUpdateDateTuple import (
    SearchIndexUpdateDateTuple,
)
from twisted.internet.defer import inlineCallbacks

logger = logging.getLogger(__name__)

clientSearchIndexUpdateFromServerFilt = dict(
    key="clientSearchIndexUpdateFromServer"
)
clientSearchIndexUpdateFromServerFilt.update(searchFilt)


class SearchIndexCacheController(ACICacheControllerABC):
    """SearchIndex Cache Controller

    The SearchIndex cache controller stores all the chunks in memory,
    allowing fast access from the mobile and desktop devices.

    """

    _ChunkedTuple = EncodedSearchIndexChunk
    _UpdateDateTupleABC = SearchIndexUpdateDateTuple
    _chunkLoadRpcMethod = ClientChunkLoadRpc.loadSearchIndexChunks
    _chunkIndexDeltaRpcMethod = ClientChunkLoadRpc.loadSearchIndexDelta
    _updateFromLogicFilt = clientSearchIndexUpdateFromServerFilt
    _logger = logger

    def __init__(self, clientId: str, pluginDataDir: Path):
        ACICacheControllerABC.__init__(self, clientId, pluginDataDir)
        self._fastKeywordController = None

    def setFastKeywordController(self, fastKeywordController):
        self._fastKeywordController = fastKeywordController

    def shutdown(self):
        ACICacheControllerABC.shutdown(self)
        self._fastKeywordController = None

    @inlineCallbacks
    def notifyFastIndexOfChunkKeysUpdated(self, chunkKeys: List[Any]):
        yield self._fastKeywordController.notifyOfUpdate(chunkKeys)



import logging
from collections import namedtuple
from datetime import datetime
from typing import List

import pytz
from sqlalchemy import select

from peek_core_search._private.storage.SearchIndex import SearchIndex
from peek_core_search._private.storage.SearchIndexCompilerQueue import (
    SearchIndexCompilerQueue,
)
from peek_core_search._private.worker.tasks.KeywordSplitter import (
    splitFullKeywords,
    splitPartialKeywords,
)
from peek_core_search._private.worker.tasks._CalcChunkKey import (
    makeSearchIndexChunkKey,
)
from peek_plugin_base.worker import CeleryDbConn

logger = logging.getLogger(__name__)

ObjectToIndexTuple = namedtuple(
    "ObjectToIndexTuple", ["id", "fullKwProps", "partialKwProps"]
)


def removeObjectIdsFromSearchIndex(deletedObjectIds: List[int]) -> None:
    pass


def reindexSearchObject(
    conn,
    objectsToIndex: List[ObjectToIndexTuple],
    excludedPartialSearchTerms: list[str],
    excludedFullSearchTerms: set[str],
) -> None:
    """Reindex Search Object

    :param excludedFullSearchTerms:
    :param excludedPartialSearchTerms: A list of strings to exclude from tokenizing
    :param conn: The database connection
    :param objectsToIndex: Object To Index
    :returns:
    """

    logger.debug("Starting to index %s SearchIndex", len(objectsToIndex))

    searchIndexTable = SearchIndex.__table__
    queueTable = SearchIndexCompilerQueue.__table__

    startTime = datetime.now(pytz.utc)

    newSearchIndexes = []
    objectIds = []
    searchIndexChunksToQueue = set()

    for objectToIndex in objectsToIndex:
        newSearchIndexes.extend(
            _indexObject(
                objectToIndex,
                excludedPartialSearchTerms,
                excludedFullSearchTerms,
            )
        )
        objectIds.append(objectToIndex.id)

    newIdGen = CeleryDbConn.prefetchDeclarativeIds(
        SearchIndex, len(newSearchIndexes)
    )
    for newSearchIndex in newSearchIndexes:
        newSearchIndex.id = next(newIdGen)
        searchIndexChunksToQueue.add(newSearchIndex.chunkKey)

    results = conn.execute(
        select(
            searchIndexTable.c.chunkKey,
        ).where(
            searchIndexTable.c.objectId.in_(objectIds),
        )
    )

    for result in results:
        searchIndexChunksToQueue.add(result.chunkKey)

    if objectIds:
        conn.execute(
            searchIndexTable.delete().where(
                searchIndexTable.c.objectId.in_(objectIds)
            )
        )

    if newSearchIndexes:
        logger.debug("Inserting %s SearchIndex", len(newSearchIndexes))
        inserts = [o.tupleToSqlaBulkInsertDict() for o in newSearchIndexes]
        conn.execute(searchIndexTable.insert(), inserts)

    if searchIndexChunksToQueue:
        conn.execute(
            queueTable.insert(),
            [dict(chunkKey=k) for k in searchIndexChunksToQueue],
        )

    logger.info(
        "Inserted %s SearchIndex keywords in %s",
        len(newSearchIndexes),
        (datetime.now(pytz.utc) - startTime),
    )


# stopwords = set()  # nltk.corpus.stopwords.words('english'))
# stopwords.update(list(string.punctuation))
#
# from nltk import PorterStemmer
#
# stemmer = PorterStemmer()


# from nltk.stem import WordNetLemmatizer
#
# lemmatizer = WordNetLemmatizer()


def _indexObject(
    objectToIndex: ObjectToIndexTuple,
    excludedPartialSearchTerms: list[str],
    excludedFullSearchTerms: set[str],
) -> List[SearchIndex]:
    """Index Object

    This method creates  the "SearchIndex" objects to insert into the DB.

    Because our data is not news articles, we can skip some of the advanced
    natural language processing (NLP)

    We're going to be indexing things like unique IDs, job titles, and equipment names.
    We may add exclusions for nuisance words later on.

    """
    searchIndexes = []

    for propKey, text in objectToIndex.fullKwProps.items():
        for token in splitFullKeywords(excludedFullSearchTerms, text):
            searchIndexes.append(
                SearchIndex(
                    chunkKey=makeSearchIndexChunkKey(token),
                    keyword=token,
                    propertyName=propKey,
                    objectId=objectToIndex.id,
                )
            )

    for propKey, text in objectToIndex.partialKwProps.items():
        for token in splitPartialKeywords(excludedPartialSearchTerms, text):
            searchIndexes.append(
                SearchIndex(
                    chunkKey=makeSearchIndexChunkKey(token),
                    keyword=token,
                    propertyName=propKey,
                    objectId=objectToIndex.id,
                )
            )

    return searchIndexes


# if __name__ == '__main__':
#     objectToIndex = ObjectToIndexTuple(
#         id=1,
#         props={
#             'key': 'COMP3453453J',
#             'alias': 'AB1345XXX',
#             'name': 'Hello, this is tokenising, strings string, child children'
#         }
#     )
#     print(_indexObject(objectToIndex))
